;;; $Id: hlco.scm,v 1.6 1994/02/04 19:41:47 queinnec Exp $

;;;                 The code of the associated paper
;;;         presenting most of the current control operators.
;;;                        Christian Queinnec
;;;           \'Ecole Polytechnique & INRIA-Rocquencourt

;;; This is the library itself. It is defined in a separate file since
;;; this code is interpreted by the interpreter defined in contlib.scm.

(define (resume q v)
  ((frame-behavior (car q)) (cdr q) v) )

(define (extend q frame)
  (cons frame q) )

(define (ACPS-call/cc q1 f)
  (f q1 (lambda (q2 value) (resume q1 value))) )

(define (extend* q frames)
  (if (pair? frames)
      (extend (extend* q (cdr frames)) (car frames))
      q ) )   

(define (split q predicate success failure)
  (define (scan q frames)
    (if (pair? q)        
        (if (predicate q)
            (success (reverse frames) q)
            (scan (cdr q) (cons (car q) frames)) )
        (failure) ) )
  (scan q '()) )

(define (ACPS-prompt-evaluate q thunk)
  (thunk (extend q (make-prompt-frame resume))) )

(define (ACPS-control q f)
  (split q (lambda (q) (prompt-frame? (car q)))
         (lambda (frames q)
           (f q (lambda (q v)
                  (resume (extend* q frames) v) )) )
         (lambda () (error "No enclosing prompt")) ) )

(define (ACPS-reset-evaluate q thunk)
  (thunk (extend q (make-reset-frame resume))) )

(define (ACPS-shift-evaluate q f)
  (split q (lambda (q) (reset-frame? (car q)))
         (lambda (frames q)
           (let ((frames (append frames (list (car q)))))
             (f q (lambda (q v)
                    (resume (extend* q frames) v) )) ) )
         (lambda () (error "No enclosing reset")) ) )

(define (ACPS-spawn q f)
  (let ((frame (make-spawn-frame resume)))
    (f (extend q frame)
       (lambda (q g)
         (split q (lambda (q) (eq? (car q) frame))
                (lambda (frames q)
                  (let ((frames (append frames (list (car q)))))
                    (g q (lambda (q v)
                           (resume (extend* q frames) v) )) ) )
                (lambda () (error "Out of extent")) ) ) ) ) )

(define (ACPS-splitter q f)
  (let* ((frame (make-splitter-frame resume))
         (q (extend q frame)) )
    (f q (make-mark q)) ) )

(define (ACPS-within-extent? q mark)
  (resume q (split q (lambda (q) (eq? q (mark-q mark)))
                   (lambda (frames q) #t)
                   (lambda () #f) )) )

(define (ACPS-abort q mark thunk)
  (split q (lambda (q) (eq? q (mark-q mark)))
         (lambda (frames q) (thunk q))
         (lambda () (error "Out of extent")) ) )

(define (ACPS-call/pc qq mark f)
  (split qq (lambda (q) (eq? q (mark-q mark)))
         (lambda (frames q)
           (f qq (lambda (q v)
                   (resume (extend* q frames) v) )) )
         (lambda () (error "Out of extent")) ) )

(define (ACPS-dynamic-wind q prelude work postlude)
  (let ((q1 (extend q (make-wind-frame
                       (lambda (q4 value)
                         (postlude
                          (extend q4
                                  (make-frame
                                   (lambda (q3 ignore-postlude-value)
                                     (resume q3 value) ) ) ) ) )
                       prelude
                       postlude ) )))
    (prelude (extend q1 (make-frame
                         (lambda (q2 ignore-prelude-value)
                           (work q2) ) ))) ) )

(define (ACPS-new-call/cc qq function)
  (define (rewind q value)
    (define (deepest-wind qq) ; find the deepest wind-frame in {\tt qq} till {\tt q}
      (if (eq? qq q)
          #f
          (or (deepest-wind (cdr qq))
              (if (wind-frame? (car qq)) qq #f) ) ) )
    (let ((wq (deepest-wind qq))) ; evaluate preludes from {\tt q} to {\tt qq}
      (if wq ((wind-frame-prelude (car wq))
              (extend (cdr wq)
                      (make-frame
                       (lambda (ignore-q ignore-prelude-value)
                         (rewind wq value) ) ) ))
          ;; resume the caller of new-call/cc
          (resume qq value) ) ) )
  (define (unwind q value)
    (define (tail? q1 q2) ; is {\tt q1} a tail of {\tt q2} ?
      (or (eq? q1 q2)
          (and (pair? q2) (tail? q1 (cdr q2))) ) )
    (if (tail? q qq) ; is {\tt q} an ancestor of {\tt qq}
        (rewind q value)
        (if (wind-frame? (car q)) ; evaluate postludes from {\tt q}
            ((wind-frame-postlude (car q))
             (extend (cdr q)
                     (make-frame
                      (lambda (ignore-q ignore-postlude-value)
                        (unwind (cdr q) value) ) ) ) )
            (unwind (cdr q) value) ) ) )
  (function qq unwind) )

;;; for my own tests

(define ACPS-call/de ACPS-splitter)

(define ACPS-within/de? ACPS-within-extent?)

;;; end of hlco.scm
